import json
from django.db import models
from hashlib import sha1


class CacheEntry(models.Model):
    service = models.CharField(max_length=50, db_index=True)
    url = models.TextField()
    url_key = models.SlugField(max_length=40, unique=True)
    status = models.PositiveIntegerField()
    content = models.TextField()
    headers = models.TextField()

    class Meta:
        app_label = 'rc_django'
        db_table = 'restclients_cacheentry'
        unique_together = ('service', 'url_key')

    def getHeaders(self):
        return json.loads(self.headers)

    def save(self, *args, **kwargs):
        self.headers = json.dumps(self.headers)
        self.url_key = self.get_url_key(self.url)
        super(CacheEntry, self).save(*args, **kwargs)

    @staticmethod
    def get_url_key(url):
        return sha1(url.encode('utf-8')).hexdigest()


class CacheEntryTimedManager(models.Manager):
    def find_nonexpired_by_service_and_url(self, service, url, time_limit):
        return super(CacheEntryTimedManager, self).get_queryset().filter(
            service=service, url_key=CacheEntry.get_url_key(url),
            time_saved__gte=time_limit)

    def find_by_service_and_url(self, service, url):
        return super(CacheEntryTimedManager, self).get_queryset().filter(
            service=service, url_key=CacheEntry.get_url_key(url))


class CacheEntryTimed(CacheEntry):
    time_saved = models.DateTimeField()

    objects = CacheEntryTimedManager()

    class Meta:
        app_label = 'rc_django'
        db_table = 'restclients_cacheentrytimed'


class DegradePerformance(object):
    def __init__(self, serialized=None):
        self.problems = {}

        if serialized:
            self.problems = json.loads(serialized)

    def remove_service(self, service):
        del self.problems[service]

    def set_status(self, service, value):
        self._set(service, "status", value)

    def set_content(self, service, value):
        self._set(service, "content", value)

    def set_load_time(self, service, value):
        self._set(service, "load_time", value)

    def get_status(self, service):
        return self._get(service, "status")

    def get_content(self, service):
        return self._get(service, "content")

    def get_load_time(self, service):
        return self._get(service, "load_time")

    def services(self):
        return self.problems.keys()

    def _set(self, service, key, value):
        self._add_service(service)
        self.problems[service][key] = value

    def _get(self, service, key):
        if service in self.problems:
            if key in self.problems[service]:
                return self.problems[service][key]
        return None

    def _add_service(self, service):
        if service not in self.problems:
            self.problems[service] = {}

    def serialize(self):
        return json.dumps(self.problems)



from restclients_core.util.performance import PerformanceDegradation
from rc_django.models import DegradePerformance


class EnableServiceDegradationMiddleware(object):
    """
    Makes it so an admin tool can set specific services to either be slower,
    have an error response code, or custom content.
    """
    # Django 1.10 MIDDLEWARE compat
    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        response = None
        if hasattr(self, 'process_request'):
            response = self.process_request(request)
        if not response:
            response = self.get_response(request)
        if hasattr(self, 'process_response'):
            response = self.process_response(request, response)
        return response

    def process_request(self, request):
        PerformanceDegradation.clear_problems()
        if "RESTCLIENTS_ERRORS" not in request.session:
            return

        problem_str = request.session["RESTCLIENTS_ERRORS"]
        problems = DegradePerformance(problem_str)
        PerformanceDegradation.set_problems(problems)

    def process_response(self, request, response):
        PerformanceDegradation.clear_problems()

        return response



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-20 20:00
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CacheEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('service', models.CharField(db_index=True, max_length=50)),
                ('url', models.TextField()),
                ('url_key', models.SlugField(max_length=40, unique=True)),
                ('status', models.PositiveIntegerField()),
                ('header_pickle', models.TextField()),
                ('content', models.TextField()),
            ],
            options={
                'db_table': 'restclients_cacheentry',
            },
        ),
        migrations.CreateModel(
            name='CacheEntryTimed',
            fields=[
                ('cacheentry_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='rc_django.CacheEntry')),
                ('time_saved', models.DateTimeField()),
            ],
            options={
                'db_table': 'restclients_cacheentrytimed',
            },
            bases=('rc_django.cacheentry',),
        ),
        migrations.AlterUniqueTogether(
            name='cacheentry',
            unique_together=set([('service', 'url_key')]),
        ),
    ]



from django.conf import settings

enable_logging = getattr(settings, 'ENABLE_MEMCACHE_LOGGING', None)


def log_err(logger, msg):
    if enable_logging:
        logger.error(msg)


def log_info(logger, msg):
    if enable_logging:
        logger.info(msg)



"""
Contains memcached cache implementations
"""

from bmemcached import Client
from bmemcached.exceptions import MemcachedException
import logging
import pickle
import threading
from dateutil.parser import parse
from django.conf import settings
from django.utils import timezone
from restclients_core.models import MockHTTP
from rc_django.cache_implementation.logger import log_err, log_info


logger = logging.getLogger(__name__)


class MemcachedCache(object):

    def __init__(self):
        self._set_client()

    def deleteCache(self, service, url):
        key = self._get_key(service, url)
        try:
            return self.client.delete(key)
        except MemcachedException as ex:
            log_err(logger, "MemCache Delete(key: {}) => {}".format(key, ex))

    def getCache(self, service, url, headers):
        key = self._get_key(service, url)
        try:
            data = self.client.get(key)
        except MemcachedException as ex:
            log_err(logger, "MemCache Get(key: {}) => {}".format(key, ex))
            return

        if not data:
            return None

        values = pickle.loads(data, encoding="utf8")
        response = MockHTTP()
        response.headers = values["headers"]
        response.status = values["status"]
        response.data = values["data"]

        return {"response": response}

    def processResponse(self, service, url, response):
        header_data = {}
        for header in response.headers:
            header_data[header] = response.getheader(header)

        key = self._get_key(service, url)
        cdata, time_to_store = self._make_cache_data(
            service, url, response.data, header_data,
            response.status, timezone.now())
        try:
            self.client.set(key, cdata, time=time_to_store)
            log_info(logger, "MemCache Set(key: {})".format(key))
        except MemcachedException as ex:
            log_err(logger, "MemCache Set(key: {}) => {}".format(key, ex))

    def updateCache(self, service, url, new_data, new_data_dt):
        """
        :param new_data: a string representation of the data
        :param new_data_dt: a timezone aware datetime object giving
                the timestamp of the new_data
        :raise MemcachedException: if update failed
        """
        key = self._get_key(service, url)
        cdata, time_to_store = self._make_cache_data(
            service, url, new_data, {}, 200, new_data_dt)
        try:
            value = self.client.get(key)
            if value:
                data = pickle.loads(value, encoding="utf8")
                if "time_stamp" in data:
                    cached_data_dt = parse(data["time_stamp"])
                    if new_data_dt <= cached_data_dt:
                        log_info(logger, "IN cache (key: {})".format(key))
                        return
                # replace existing value in cache
                self.client.replace(key, cdata, time=time_to_store)
                log_info(logger, "MemCache replace(key: {})".format(key))
                return
        except MemcachedException as ex:
            log_err(logger, "MemCache replace(key: {}) => {}".format(key, ex))
            raise

        # not in cache
        try:
            self.client.set(key, cdata, time=time_to_store)
            log_info(logger, "MemCache Set(key {})".format(key))
        except MemcachedException as ex:
            log_err(logger, "MemCache Set(key: {}) => {}".format(key, ex))
            raise

    def _make_cache_data(self, service, url, data_to_cache,
                         header_data, status, time_stamp):
        data = {"status": status,
                "headers": header_data,
                "data": data_to_cache,
                "time_stamp": time_stamp.isoformat()
                }
        time_to_store = self.get_cache_expiration_time(service, url)
        return pickle.dumps(data), time_to_store

    def get_cache_expiration_time(self, service, url):
        # Over-ride this to define your own.
        return 60 * 60 * 4

    def _get_key(self, service, url):
        return "{}-{}".format(service, url)

    def _set_client(self):
        thread_id = threading.current_thread().ident
        if not hasattr(MemcachedCache, "_memcached_cache"):
            MemcachedCache._memcached_cache = {}

        if thread_id in MemcachedCache._memcached_cache:
            self.client = MemcachedCache._memcached_cache[thread_id]
            return

        servers = settings.RESTCLIENTS_MEMCACHED_SERVERS
        username = getattr(settings, "RESTCLIENTS_MEMCACHED_USER", None)
        password = getattr(settings, "RESTCLIENTS_MEMCACHED_PASS", None)

        self.client = Client(servers, username, password)
        MemcachedCache._memcached_cache[thread_id] = self.client



"""
Contains DAO Cache implementations
"""

import json
from logging import getLogger
import threading
from base64 import b64encode, b64decode
from datetime import datetime, timedelta
from django.conf import settings
from django.utils.timezone import make_aware, get_current_timezone
from restclients_core.cache_manager import store_cache_entry
from restclients_core.models import MockHTTP
from rc_django.models import CacheEntryTimed


logger = getLogger(__name__)


class TimedCache(object):
    """
    This is a base class for Cache implementations that caches for
    lengths of time.
    """
    def deleteCache(self, service, url):
        rs = CacheEntryTimed.objects.find_by_service_and_url(service, url)
        if len(rs):
            return rs.delete()
        return None

    def _response_from_cache(self, service, url, headers, max_age_in_seconds,
                             max_error_age=60 * 5):

        # If max_age_in_seconds is 0,
        # make sure we don't get a hit from this same second.
        if not max_age_in_seconds:
            return None
        now = make_aware(datetime.now(), get_current_timezone())
        time_limit = now - timedelta(seconds=max_age_in_seconds)

        query = CacheEntryTimed.objects.find_nonexpired_by_service_and_url(
            service, url, time_limit)

        if len(query):
            hit = query[0]

            if hit.status != 200 and (
                    now - timedelta(seconds=max_error_age) > hit.time_saved):
                return None

            response = MockHTTP()
            response.status = hit.status
            response.data = hit.content
            response.headers = hit.getHeaders()

            return {
                "response": response,
            }
        return None

    def _process_response(self, service, url, response,
                          overwrite_success_with_error_at=60 * 60 * 8):
        now = make_aware(datetime.now(), get_current_timezone())
        query = CacheEntryTimed.objects.find_by_service_and_url(service, url)

        cache_entry = None
        if len(query):
            cache_entry = query[0]
        else:
            cache_entry = CacheEntryTimed()

        if response.status != 200:
            # Only override a successful cache entry with an error if the
            # Successful entry is older than 8 hours - MUWM-509
            if cache_entry.id is not None and cache_entry.status == 200:
                save_delta = now - cache_entry.time_saved
                extended_cache_delta = timedelta(
                    seconds=overwrite_success_with_error_at)

                if save_delta < extended_cache_delta:
                    response = MockHTTP()
                    response.status = cache_entry.status
                    response.data = cache_entry.content
                    return {"response": response}

        cache_entry.service = service
        cache_entry.url = url
        cache_entry.status = response.status
        try:
            cache_entry.content = response.data.decode('utf-8')
        except AttributeError:
            cache_entry.content = response.data

        # This extra step is needed w/ Live resources because
        # HTTPHeaderDict isn't serializable.
        header_data = {}
        if response.headers is not None:
            for header in response.headers:
                header_data[header] = response.getheader(header)

        cache_entry.headers = header_data
        cache_entry.time_saved = now

        try:
            store_cache_entry(cache_entry)
        except Exception as ex:
            # If someone beat us in to saving a cache entry, that's ok.
            # We just need a very recent entry.
            return

        return


class TimeSimpleCache(TimedCache):
    """
    This caches all URLs for 60 seconds.  Used for testing.
    """
    def getCache(self, service, url, headers):
        return self._response_from_cache(service, url, headers, 60)

    def processResponse(self, service, url, response):
        return self._process_response(service, url, response)


class FourHourCache(TimedCache):
    """
    This caches all URLs for 4 hours.  Provides a basic way to cache
    cache resources that don't give a useful expires header, but you don't
    want to make a server round trip to validate an etag for.
    """
    def getCache(self, service, url, headers):
        return self._response_from_cache(service, url, headers,  60 * 60 * 4)

    def processResponse(self, service, url, response):
        return self._process_response(service, url, response)


class ETagCache(object):
    """
    This caches objects just based on ETags.
    """
    def getCache(self, service, url, headers):
        now = make_aware(datetime.now(), get_current_timezone())
        time_limit = now - timedelta(seconds=60)

        query = CacheEntryTimed.objects.find_by_service_and_url(service, url)

        if len(query):
            hit = query[0]

            response = MockHTTP()
            response.status = hit.status
            response.data = hit.content

            hit_headers = hit.getHeaders()

            if "ETag" in hit_headers:
                headers["If-None-Match"] = hit_headers["ETag"]

        return None

    def processResponse(self, service, url, response):
        query = CacheEntryTimed.objects.find_by_service_and_url(service, url)

        cache_entry = CacheEntryTimed()
        if len(query):
            cache_entry = query[0]

        if response.status == 304:
            if cache_entry is None:
                raise Exception("304, but no content??")

            response = MockHTTP()
            response.status = cache_entry.status
            response.data = cache_entry.content
            response.headers = cache_entry.headers
            return {"response": response}
        else:
            now = make_aware(datetime.now(), get_current_timezone())
            cache_entry.service = service
            cache_entry.url = url
            cache_entry.status = response.status
            cache_entry.content = response.data

            cache_entry.headers = response.headers
            cache_entry.time_saved = now
            store_cache_entry(cache_entry)

        return



from django.test import TestCase
from restclients_core.dao import DAO, MockDAO
from restclients_core.models import MockHTTP
from rc_django.cache_implementation import TimeSimpleCache, FourHourCache
from rc_django.models import CacheEntryTimed
from datetime import timedelta


class TIMED_DAO(DAO):
    def service_name(self):
        return "timed"

    def get_default_service_setting(self, key):
        if "DAO_CLASS" == key:
            return "rc_django.tests.cache_implementation.test_timed.Backend"


class SIMPLE_DAO(TIMED_DAO):
    def service_name(self):
        return "simple"

    def get_cache(self):
        return TimeSimpleCache()


class FOUR_DAO(TIMED_DAO):
    def service_name(self):
        return "four"

    def get_cache(self):
        return FourHourCache()


class Backend(MockDAO):
    def load(self, method, url, headers, body):
        response = MockHTTP()
        if url == "/same":
            response.status = 200
            response.data = b"Body Content"
        else:
            response.status = 404
        return response


class TimeCacheTest(TestCase):
    def test_saved_headers(self):
        cache = TimeSimpleCache()

        response = MockHTTP()
        response.status = 200
        response.data = "Cache testing"
        response.headers = {
            "link": "next,http://somewhere",
            "Content-type": "text",
            "random": "stuff",
            "and": "more",
        }

        cache._process_response("cache_test", "/v1/headers", response)

        from_db = cache._response_from_cache(
            "cache_test", "/v1/headers", {}, 10)

        self.assertEquals(from_db["response"].status, 200)
        self.assertEquals(from_db["response"].data, "Cache testing")
        self.assertEquals(from_db["response"].getheader("random"), "stuff")
        self.assertEquals(from_db["response"].getheader("and"), "more")
        self.assertEquals(
            from_db["response"].getheader("Content-type"), "text")
        self.assertEquals(
            from_db["response"].getheader("link"), "next,http://somewhere")

    def test_simple_cache(self):
        # Check initial state
        cache = TimeSimpleCache()
        response = cache.getCache('simple', '/same', {})
        self.assertEquals(response, None)
        self.assertIsNone(cache.deleteCache('simple', '/same'))

        client = SIMPLE_DAO()
        response = client.getURL('/same', {})

        self.assertEquals(response.status, 200)
        self.assertEquals(response.data, b"Body Content")

        # Make sure there's a response there after the get
        hit = cache.getCache('simple', '/same', {})
        response = hit["response"]

        self.assertEquals(response.status, 200)
        self.assertEquals(response.data, "Body Content")

        self.assertIsNotNone(cache.deleteCache('simple', '/same'))

    def test_4hour_cache(self):
        # Check initial state
        cache = FourHourCache()
        response = cache.getCache('four', '/same', {})
        self.assertEquals(response, None)

        client = FOUR_DAO()
        response = client.getURL('/same', {})

        self.assertEquals(response.data, b"Body Content")

        # Make sure there's a response there after the get
        hit = cache.getCache('four', '/same', {})
        response = hit["response"]

        self.assertEquals(response.status, 200)
        self.assertEquals(response.data, "Body Content")

        cache_entry = CacheEntryTimed.objects.get(service="four", url="/same")

        # Cached response is returned after 3 hours and 58 minutes
        orig_time_saved = cache_entry.time_saved
        cache_entry.time_saved = orig_time_saved - timedelta(minutes=238)
        cache_entry.save()

        hit = cache.getCache('four', '/same', {})
        self.assertNotEquals(hit, None)

        # Cached response is not returned after 4 hours and 1 minute
        cache_entry.time_saved = orig_time_saved - timedelta(hours=241)
        cache_entry.save()

        hit = cache.getCache('four', '/same', {})
        self.assertEquals(hit, None)

    def test_errors(self):
        cache = FourHourCache()
        response = cache.getCache('four', '/invalid', {})
        self.assertEquals(response, None)

        client = FOUR_DAO()
        response = client.getURL('/invalid', {})

        hit = cache.getCache('four', '/invalid', {})
        response = hit["response"]

        self.assertEquals(response.status, 404)

        cache_entry = CacheEntryTimed.objects.get(
            service="four", url="/invalid")

        # Make sure that invalid entry stops being returned after 5 mintes
        cache_entry.time_saved = cache_entry.time_saved - timedelta(minutes=5)
        cache_entry.save()

        hit = cache.getCache('four', '/invalid', {})
        self.assertEquals(hit, None, "No hit on old, bad status codes")

        # Make sure bad responses don't overwrite good ones.
        ok_response = MockHTTP()
        ok_response.status = 200
        ok_response.data = "OK"

        cache.processResponse("four", "/ok", ok_response)

        cache_response = cache.getCache("four", "/ok", {})
        response = cache_response["response"]
        self.assertEquals(response.status, 200)

        bad_response = MockHTTP()
        bad_response.status = 500
        bad_response.data = "This is bad data"

        cache.processResponse("four", "/ok", bad_response)
        cache_response = cache.getCache("four", "/ok", {})
        response = cache_response["response"]
        self.assertEquals(response.status, 200)
        self.assertEquals(response.data, "OK")

        # Make sure that an old, good hit is returned when there's a fresh,
        # bad hit.
        ok_response = MockHTTP()
        ok_response.status = 200
        ok_response.data = "Valid"

        cache.processResponse("four", "/valid", ok_response)

        response = client.getURL("/valid", {})
        self.assertEquals(response.status, 200)

        cache_entry = CacheEntryTimed.objects.get(
            service="four", url="/valid")

        cache_entry.time_saved = cache_entry.time_saved - timedelta(hours=5)
        cache_entry.save()

        response = client.getURL("/valid", {})
        self.assertEquals(response.status, 200)

        # But make sure eventually we stop using our cache.
        cache_entry.time_saved = cache_entry.time_saved - timedelta(hours=9)
        cache_entry.save()

        response = client.getURL("/valid", {})
        self.assertEquals(response.status, 404)



import json
from unittest import skipIf
from django.conf import settings
from django.test import TestCase
from django.utils import timezone
from restclients_core.models import MockHTTP
from rc_django.cache_implementation.memcache import (
    MemcachedCache, Client, MemcachedException)


@skipIf(not getattr(settings, 'RESTCLIENTS_MEMCACHED_SERVERS', None),
        "Memcached cache not configured")
class MemcachedCacheTest(TestCase):
    def setUp(self):
        cache = MemcachedCache()
        cache.client.flush_all()

    def test_cacheGet(self):
        cache = MemcachedCache()
        key = cache._get_key('mem', '/same')
        self.assertEquals(key, "mem-/same")

        data = cache.getCache('mem', '/same', {})
        self.assertIsNone(data)

        cdata, time_to_store = cache._make_cache_data(
            'mem', '/same', json.dumps({"data": "Body Content"}),
            {}, 200, timezone.now())
        cache.client.set(key, cdata, time_to_store)

        hit = cache.getCache('mem', '/same', {})
        response = hit["response"]
        self.assertEquals(response.status, 200)
        self.assertEquals(response.data, '{"data": "Body Content"}')

        # delete existing entry
        self.assertTrue(cache.deleteCache('mem', '/same'))

        cache.client = MockClient1()
        # test get err
        cache.getCache('mem', '/same', {})
        # test delete err
        self.assertFalse(cache.deleteCache('mem', '/same'))

    def test_updateCache(self):
        cache = MemcachedCache()
        # cache no data
        cache.updateCache('mem', '/same', '{"data": "Content1"}',
                          timezone.now())

        time1 = timezone.now()
        # cache has older data
        cache.updateCache('mem', '/same', '{"data": "Content2"}', time1)

        hit = cache.getCache('mem', '/same', {})
        response = hit["response"]
        self.assertEquals(response.headers, {})
        self.assertEquals(response.status, 200)
        self.assertEquals(response.data, '{"data": "Content2"}')

        # update with no newer data
        cache.updateCache('mem', '/same', '{"data": "Content3"}', time1)
        hit = cache.getCache('mem', '/same', {})
        response = hit["response"]
        self.assertEquals(response.data, '{"data": "Content2"}')

        # test replace err
        cache.client = MockClient2()
        self.assertRaises(MemcachedException,
                          cache.updateCache,
                          'mem', '/same', '{}', timezone.now())

        # test set err
        cache.deleteCache('mem', '/same')
        self.assertRaises(MemcachedException,
                          cache.updateCache,
                          'mem', '/same', '{}', timezone.now())

    def test_processResponse(self):
        mock_resp = MockHTTP()
        mock_resp.status = 200
        mock_resp.data = "Content4"
        mock_resp.headers = {"Content-type": "text/html"}

        cache = MemcachedCache()
        cache.processResponse('mem', '/same1', mock_resp)

        hit = cache.getCache('mem', '/same1', {})
        response = hit["response"]
        self.assertEquals(response.headers, {"Content-type": "text/html"})
        self.assertEquals(response.status, 200)
        self.assertEquals(response.data, "Content4")

    def test_processResponse_err(self):
        mock_resp = MockHTTP()
        mock_resp.status = 200
        mock_resp.data = "Content4"
        mock_resp.headers = {"Content-type": "text/html"}
        cache = MemcachedCache()
        cache.client = MockClient2()
        cache.processResponse('mem', '/same1', mock_resp)

    def test_binary_processResponse(self):
        mock_resp = MockHTTP()
        mock_resp.status = 200
        mock_resp.data = b'content to be encoded'
        mock_resp.headers = {"Content-type": "image/png"}

        cache = MemcachedCache()
        cache.processResponse('mem', '/same2', mock_resp)

        hit = cache.getCache('mem', '/same2', {})
        response = hit["response"]
        self.assertEquals(response.headers, {"Content-type": "image/png"})
        self.assertEquals(response.status, 200)
        self.assertEquals(response.data, b'content to be encoded')

    def test_memcached_client(self):
        cache = MemcachedCache()
        key = cache._get_key('mem', '/same')
        self.assertEquals(key, "mem-/same")
        cache.updateCache('mem', '/same', '{"data": "Content"}',
                          timezone.now())
        hit = cache.getCache('mem', '/same', {})
        response = hit["response"]
        self.assertEquals(response.headers, {})
        self.assertEquals(response.status, 200)


class MockClient1(Client):

    def delete(self, key):
        raise MemcachedException("err", 400)

    def get(self, key, get_cas=False):
        raise MemcachedException("err", 400)


class MockClient2(Client):

    def replace(self, key, value, time=0, compress_level=-1):
        raise MemcachedException("err", 400)

    def set(self, key, value, time=0, compress_level=-1):
        raise MemcachedException("err", 400)
